from interactions import mono_assign
from particles import enc_mat
from tqdm import tqdm
from functools import partial
import multiprocessing as mp


def nested_assign(x, hh, ww, stride, new_w, new_h, c, ij):
    out_mat = ['' for t in range(c * hh * ww)]
    i = ij // new_w
    j = ij // new_h
    for cidx in range(c):
        for a in range(hh):
            for b in range(ww):
                out_mat[cidx * hh * ww + a * ww + b] =\
                    x[cidx][i * stride + a][j * stride + b]
    return out_mat


def im2col(name, x, hh, ww, stride=1):
    """
    Args:
      x: image matrix to be translated into columns, (C,H,W)
      hh: filter height
      ww: filter width
      stride: stride
    Returns:
      col: (new_h*new_w,hh*ww*C) matrix, each column is a
           cube that will convolve with a filter
           new_h = (H-hh) // stride + 1, new_w = (W-ww) // stride + 1
    """

    c, h, w = x.size[0], x.size[1], x.size[2]
    new_h = (h - hh) // stride + 1
    new_w = (w - ww) // stride + 1
    col = enc_mat(name=name + 'i2' + str(c * h * w),
                  size=(new_h * new_w, c * hh * ww))
    old_lst_ = x._lst
    old_name_lst_ = x.name_list
    p = mp.Pool(mp.cpu_count())
    new_lst_ = p.map(partial(nested_assign,
                             old_lst_,
                             hh,
                             ww,
                             stride,
                             new_w,
                             new_h,
                             c),
                     range(new_h * new_w))
    p.close()
    p.join()
    p = mp.Pool(mp.cpu_count())
    new_name_lst_ = p.map(partial(nested_assign,
                                  old_name_lst_,
                                  hh,
                                  ww,
                                  stride,
                                  new_w,
                                  new_h,
                                  c),
                          range(new_h * new_w))
    col.reorder(new_lst_, new_name_lst_)
    p.close()
    p.join()
    return col


def col2im(circuit, mul, h_prime, w_prime, C, tgt_out):
    """
      Args:
      mul: (h_prime*w_prime*w,F) matrix,
           each col should be reshaped to
           C*h_prime*w_prime when C>0, or h_prime*w_prime when C = 0
      h_prime: reshaped filter height
      w_prime: reshaped filter width
      C: reshaped filter channel, if 0, reshape the filter to 2D,
         Otherwise reshape it to 3D
    Returns:
      if C == 0: (F,h_prime,w_prime) matrix
      Otherwise: (F,C,h_prime,w_prime) matrix
    """
    F = mul.size[1]
    for i in range(0, F):
        map(lambda idx: circuit.add(
            mono_assign(ass_type='alias',
                        lhs=[mul[idx][i]],
                        rhs=tgt_out[i][idx
                                       / w_prime][idx % w_prime])),
            range(w_prime * h_prime))



from reusable_modules import oneb_adder
from particles import variables
from utils import parse_test_output
import os
import subprocess

if "SHEEP_HOME" in os.environ.keys():
    BASE_DIR = os.environ["SHEEP_HOME"]
else:
    BASE_DIR = os.environ["HOME"] + "/SHEEP"

CIRCUIT_DIR_MID = BASE_DIR + "/benchmark_inputs/mid_level/circuits/"
INPUTS_DIR_MID = BASE_DIR + "/benchmark_inputs/mid_level/inputs/TMP/"
EXECUTABLE_DIR = BASE_DIR + "/build/bin"


class circuit(object):
    def __init__(self, name, circuit,
                 const_inputs=[]):
        self.name = name

        self.circuit = circuit
        self.const_inputs = const_inputs
        self.flag_inputs = ['TRUE', 'FALSE']
        self.on_file = False

    def get_inputs(self):
        assert self.circuit is not None
        self.const_inputs = self.const_inputs  # + self.circuit.const_inputs
        self.inputs = self.circuit.inputs
        self.outputs = self.circuit.outputs
        self.num_inputs = len(self.circuit.inputs)

    def output_circuit(self, filename):
        with open(filename, 'wb') as outfile:
            outfile.write("CONST_INPUTS")
            outfile.write("\n")
            outfile.write("INPUTS ")
            for ci in self.flag_inputs:
                outfile.write(" " + ci)
            for i in self.inputs.get_variables():
                outfile.write(" " + i)
            outfile.write("\n")
            outfile.write("OUTPUTS ")
            for o in self.outputs.get_variables():
                outfile.write(" " + o)
            outfile.write("\n")
            outfile.write(str(self.circuit))
            outfile.write("\n")

    def write_file(self, filename=None):
        self.get_inputs()
        self.on_file = True
        if filename is None:
            self.write_default_file()
        else:
            self.write_spec_file(filename)

    def write_default_file(self):
        filename = CIRCUIT_DIR_MID + self.name + ".sheep"
        self.filename = filename
        self.output_circuit(filename=filename)

    def write_spec_file(self, filename):
        self.filename = filename
        self.output_circuit(filename=filename)

    def run_circuit(self, inputs_file, input_type='bool', context='TFHE',
                    eval_strategy="parallel",
                    params_file=None, debugfilename=None):
        """
        run the circuit and retrieve the results.
        """
        run_cmd = []
        assert self.on_file, "Circuit needs to be written to file"
        run_cmd.append(os.path.join(EXECUTABLE_DIR, "benchmark"))
        run_cmd.append(self.filename)
        run_cmd.append(context)
        run_cmd.append(input_type)
        run_cmd.append(inputs_file)
        run_cmd.append(eval_strategy)
        if params_file:
            run_cmd.append(params_file)
        p = subprocess.Popen(args=run_cmd, stdout=subprocess.PIPE)
        job_output = p.communicate()[0]
        results = parse_test_output(job_output, debugfilename)
        # now just add other fields into the "results" dict, to go into the db
        results["context"] = context
        results["circuit_name"] = self.name
        results["num_inputs"] = self.num_inputs
        return results


if __name__ == '__main__':
    x = variables('x')
    y = variables('y')
    cin = variables('cin')
    s = variables('sum')
    c = variables('carry')
    b2_adder = oneb_adder('2b_adder', [x, y, cin], [s, c], 1)
    circuit = circuit('2 Bit Adder', b2_adder)
    circuit.write_file(filename='./test.sheep')



from particles import enc_vec, enc_mat, enc_tensor3
from interactions import mini_mod, mono_assign


class vec_elemwise(mini_mod):
    def __init__(self, name, inputs, outputs):
        mini_mod.__init__(self, name=name, inputs=inputs, outputs=outputs)
        assert(isinstance(inputs, enc_vec))
        assert(isinstance(outputs, enc_vec))
        assert len(self.inputs) == len(self.outputs)
        self.size = len(inputs)

    def operate(self, **kwargs):
        pass


class vec_mono_op(vec_elemwise):
    def __init__(self, ass_type, name, inputs, outputs):
        vec_elemwise.__init__(self, name=name, inputs=inputs, outputs=outputs)
        self.ass_type = ass_type
        self.size = len(self.inputs)
        self.operate()

    def operate(self):
        for idx in range(self.size):
            self.add(ass_type=self.ass_type,
                     lhs=[self.inputs[idx]],
                     rhs=self.outputs[idx])


class vec_mono_op_cond(vec_elemwise):
    def __init__(self, cond, ass_types, name, inputs, outputs):
        vec_elemwise.__init__(self, name=name, inputs=inputs, outputs=outputs)
        self.ass_types = ass_types
        self.cond = cond
        self.operate()

    def operate(self):
        for idx in range(self.size):
            self.add(mono_assign(ass_type=self.ass_types[self.cond[idx]],
                                 lhs=[self.inputs[idx]],
                                 rhs=self.outputs[idx]))


class mat_elemwise(mini_mod):
    def __init__(self, name, inputs, outputs):
        mini_mod.__init__(self, name=name, inputs=inputs, outputs=outputs)
        assert(isinstance(inputs, enc_mat))
        assert(isinstance(outputs, enc_mat))
        assert len(self.inputs) == len(self.outputs)
        assert(len(self.inputs[0])) == len(self.outputs[0])
        self.size = inputs.size

    def operate(self, **kwargs):
        pass


class mat_mono_op(vec_elemwise):
    def __init__(self, ass_type, name, inputs, outputs):
        vec_elemwise.__init__(self, name=name, inputs=inputs, outputs=outputs)
        self.ass_type = ass_type
        self.operate()

    def operate(self):
        for idx in range(self.size[0]):
            for jdx in range(self.size[1]):
                self.add(ass_type=self.ass_type,
                         lhs=[self.inputs[idx][jdx]],
                         rhs=self.outputs[idx][jdx])


class mat_mono_op_cond(vec_elemwise):
    def __init__(self, cond, ass_types, name, inputs, outputs):
        vec_elemwise.__init__(self, name=name, inputs=inputs, outputs=outputs)
        self.ass_types = ass_types
        self.cond = cond
        self.operate()

    def operate(self):
        for idx in range(self.size[0]):
            for jdx in range(self.size[1]):
                self.add(
                    mono_assign(ass_type=self.ass_types[self.cond[idx][jdx]],
                                lhs=[self.inputs[idx][jdx]],
                                rhs=self.outputs[idx][jdx]))


class tensor3_elemwise(mini_mod):
    def __init__(self, name, inputs, outputs):
        mini_mod.__init__(self, name=name, inputs=inputs, outputs=outputs)
        assert(isinstance(inputs, enc_tensor3))
        assert(isinstance(outputs, enc_tensor3))
        assert len(self.inputs) == len(self.outputs)
        assert len(self.inputs[0]) == len(self.outputs[0])
        assert len(self.inputs[0][0]) == len(self.outputs[0][0])
        self.size = self.inputs.size

    def operate(self, **kwargs):
        pass


class tensor3_mono_op(vec_elemwise):
    def __init__(self, ass_type, name, inputs, outputs):
        vec_elemwise.__init__(self, name=name, inputs=inputs, outputs=outputs)
        self.ass_type = ass_type
        self.operate()

    def operate(self):
        for idx in range(self.size[0]):
            for jdx in range(self.size[1]):
                for kdx in range(self.size[2]):
                    self.add(ass_type=self.ass_type,
                             lhs=[self.inputs[idx][jdx][kdx]],
                             rhs=self.outputs[idx][jdx][kdx])


class tensor3_mono_op_cond(vec_elemwise):
    def __init__(self, cond, ass_types, name, inputs, outputs):
        vec_elemwise.__init__(self, name=name, inputs=inputs, outputs=outputs)
        self.ass_types = ass_types
        self.cond = cond
        self.operate()

    def operate(self):
        for idx in range(len(self.inputs)):
            self.add(mono_assign(ass_type=self.ass_types[self.cond[idx]],
                                 lhs=[self.inputs[idx]],
                                 rhs=self.outputs[idx]))



from particles import enc_vec
from reusable_modules import nb_adder_xy
from interactions import mini_mod, mono_assign
import math


class reduce_add(mini_mod):
    def __init__(self, name, inputs, outputs):
        '''
        Inputs :
        ================================
        inp_bits : Input Vector , type  (enc_vec)
        func : Function for reduce
               Must take two n bit enc_vec and output a k bit enc_vec
        out_bits : Output Vector
        ==================================
        Output a circuit for reduce function as a tree
        '''
        mini_mod.__init__(self, name=name, inputs=inputs, outputs=outputs)
        self.nb = len(self.inputs)
        self.create()

    def create(self):
        inp_vec = self.inputs
        out = self.count_arr(
            lvl=int(math.ceil(math.log(len(inp_vec) * 1.0, 2))))
        for idx in range(len(self.outputs)):
            self.add(mono_assign(ass_type='alias',
                                 lhs=[out[idx]],
                                 rhs=self.outputs[idx]))

    def count_arr(self, lvl):
        inp_len = len(self.inputs)
        lvl_idx = 1
        inp_vec = self.inputs
        while(lvl_idx <= lvl):
            num_parts = int(math.ceil(inp_len * 1.0 / lvl_idx))
            left_out = []
            if num_parts % 2 == 1:
                left_out = inp_vec[(num_parts - 1) * (lvl_idx):]
                num_parts -= 1

            temp_out = enc_vec(name=self.name + '_red_out_' + str(lvl_idx),
                               nb=num_parts / 2 * (lvl_idx + 1),
                               randomize_name=10)
            for part_id in range(num_parts / 2):

                relevant_input = inp_vec[part_id * (2 * lvl_idx):
                                         (part_id + 1) * (2 * lvl_idx)]
                self.sum_part(
                    (relevant_input[0:lvl_idx], relevant_input[lvl_idx:]),
                    temp_out[part_id * (lvl_idx + 1):
                             (part_id + 1) * (lvl_idx + 1)],
                    part_id=part_id)

            if len(left_out) == 0:
                inp_vec = temp_out
            else:
                inp_vec = temp_out + left_out
            inp_len = len(inp_vec)
            lvl_idx += 1

        return inp_vec

    def sum_part(self, inp_vec, out_vec, part_id, carry=False):
        '''
        Calculates the binary encoding
        of the number of bits in the input vector.
        uses out_vec to output the results.
        '''
        self.add(nb_adder_xy(name=out_vec.name + '_adder_' + str(part_id),
                             inputs=inp_vec,
                             outputs=out_vec,
                             nb=len(inp_vec[0])))


if __name__ == '__main__':
    inp = enc_vec(name='in', nb=4)
    out = enc_vec(name='out', nb=3)
    ra = reduce_add('adder', inp, out)
    print(ra)



from vector_ops import vec_mono_op_cond
from interactions import mini_mod, bi_assign
from functools import partial
from particles import enc_vec, enc_mat, variables
from functions import reduce_add
from reusable_modules import compare_cp
from op_utils import col2im, im2col
import numpy as np
import math
import multiprocessing as mp
import time


def create_ith_ll(name_, filter_col_, im_col_, out_, map_inp):
    col_idx = map_inp[1]
    temp_lyr = linear_layer(name=name_ + 'ln' + str(col_idx),
                            weight=filter_col_,
                            inputs=im_col_[col_idx],
                            outputs=out_[col_idx], parallel=False)
    map_inp[0].put(temp_lyr)
    return 1


class nn_layer(mini_mod):

    def __init__(self, layer_type, weight, name, inputs, outputs,
                 parallel=False):
        mini_mod.__init__(self, name=name, inputs=inputs,
                          outputs=outputs, parallel=parallel)
        self.weight = weight
        isinstance(self.weight, np.ndarray)
        self.check_weight_validity()

    def check_weight_validity(self):
        pass

    def create(self):
        pass


class sign_fn(nn_layer):
    def __init__(self, name,  inputs, outputs, input_length, weight=None):
        nn_layer.__init__(self, name=name, inputs=inputs, outputs=outputs,
                          weight=None, layer_type=name + '_sign_')
        self.input_length = input_length
        self.create()

    def check_weight_validity(self):
        '''
        It is not parameteried for now
        But can include bias here later
        '''
        assert self.weight is None, "Sign functions is not parameterized"

    def create(self):
        if isinstance(self.outputs, enc_vec):
            assert(len(self.outputs) == 1), "Only one bit output"
        else:
            assert(isinstance(self.outputs, variables)), "Only one bit output"
        # self.add(bi_assign(ass_type='or', lhs=[
        #         self.inputs[-1], self.inputs[-2]],
        #    rhs=self.outputs[0]))
        self.add(compare_cp(self.name + 'cpg_',
                            inputs=(self.inputs, self.input_length / 2),
                            outputs=self.outputs))


class linear_layer_1d(nn_layer):
    def __init__(self, name, weight, inputs, outputs):
        nn_layer.__init__(self, name=name, inputs=inputs,
                          outputs=outputs, weight=weight,
                          layer_type='l_1d')
        self.nb = weight.size
        self.create()

    def check_weight_validity(self):
        '''
        Only supports vector ops for now.
        '''
        assert(self.weight.ndim == 1)
        assert(self.weight.size == len(self.inputs)), str(
            self.weight.size) + ' ~ ' + str(len(self.inputs))

    def create(self):
        _tmp = enc_vec(nb=len(self.inputs), name=self.name + '_xnor_tmp_')
        self.add(vec_mono_op_cond(name=self.name,
                                  cond=self.weight,
                                  ass_types=('not', 'alias'),
                                  inputs=self.inputs,
                                  outputs=_tmp))
        depth = int(math.floor(math.log(self.nb, 2)))
        _tmp_out = enc_vec(nb=depth + 1, name=self.name + '_mult_out')

        self.add(reduce_add(name=self.name + '_ra1_',
                            inputs=_tmp, outputs=_tmp_out))
        self.add(sign_fn(name=self.name + 'sgn',
                         inputs=_tmp_out, outputs=self.outputs,
                         input_length=len(self.inputs)))


class linear_layer(nn_layer):
    def __init__(self, name, weight, inputs, outputs, parallel=False):
        nn_layer.__init__(self, name=name, inputs=inputs,
                          outputs=outputs, weight=weight,
                          layer_type='linear', parallel=parallel)
        self.nb = weight[0].size
        self.create()

    def check_weight_validity(self):
        '''
        Only supports vector ops for now.
        '''
        assert(self.weight.ndim == 2)
        if isinstance(self.inputs, enc_mat):
            assert(self.weight.shape[1] == len(self.inputs[0])), str(
                self.weight.shape[1]) + ' ~ ' + str(len(self.inputs[0]))
        else:
            assert(self.weight.shape[1] == len(self.inputs)), str(
                self.weight.shape[1]) + ' ~ ' + str(len(self.inputs))
        assert(self.weight.shape[0] == len(self.outputs)), str(
            self.weight.shape[0]) + " ~ " + str(len(self.outputs))

    def create(self):
        if isinstance(self.inputs, enc_mat):
            rel_inp = self.inputs[0]
        else:
            rel_inp = self.inputs

        for idx in range(len(self.weight)):
            self.add(linear_layer_1d(name=self.name + 'lin_lyr' + str(idx),
                                     weight=self.weight[idx],
                                     inputs=rel_inp,
                                     outputs=self.outputs[idx]))


class conv_layer(nn_layer):
    def __init__(self, name,
                 weight, inputs, outputs,
                 stride=1, padding=0):
        self.padding = padding
        self.stride = stride
        nn_layer.__init__(self, name=name, inputs=inputs, outputs=outputs,
                          weight=weight, layer_type='conv', parallel=False)
        self.nb = reduce(lambda x, y: x * y, self.inputs.size)
        self.create()

    def check_weight_validity(self):
        assert(self.weight.ndim == 4)
        assert(self.padding == 0), "Only supports zero padding"

    def create(self):
        """
        A naive implementation of the forward pass for a convolutional layer.

        The input consists of N data points, each with C channels,
        height H and width W. We convolve each input with F different filters,
        where each filter spans  all C channels and has height HH and width HH.

        Input:
        - x: Input data of shape (N, C, H, W)
        - w: Filter weights of shape (F, C, HH, WW)
        - b: Biases, of shape (F,)
        - conv_param: A dictionary with the following keys:
        - 'stride': The number of pixels between adjacent receptive
        fields in the horizontal and vertical directions.
        - 'pad': The number of pixels that will be used to zero-pad the input.

        Returns a tuple of:
        - out: Output data, of shape (N, F, H', W') where H' and W' are
        H' = 1 + (H + 2 * pad - HH) / stride
        W' = 1 + (W + 2 * pad - WW) / stride
        - cache: (x, w, b, conv_param)
        """

        pad_num = self.padding
        stride = self.stride
        C, H, W = self.inputs.size
        F, C, HH, WW = self.weight.shape
        H_prime = (H + 2 * pad_num - HH) // stride + 1
        W_prime = (W + 2 * pad_num - WW) // stride + 1
        out = enc_mat(name=self.name + 'c_i',
                      size=(H_prime * W_prime, F))

        #print("Entering im2col....")
        beg = time.time()
        im_col = im2col(name=self.name + 'im2col',
                        x=self.inputs, hh=HH, ww=WW, stride=stride)
        end = time.time()
        #print("Exiting im2col after " + str(end - beg) + str("sec."))
        # print()
        #print("Adding linear layers")
        beg = time.time()
        filter_col = np.reshape(self.weight, (F, -1))
        cpus = mp.cpu_count()
        p = mp.Pool(cpus)
        queue = mp.Manager().Queue()
        p.map(partial(create_ith_ll,
                      self.name,
                      filter_col,
                      im_col,
                      out),
              [(queue, x) for x in range(im_col.size[0])])
        p.close()
        p.join()
        for col_idx in range(im_col.size[0]):
            self.add(queue.get())
        end = time.time()
        #print("Exiting im2col after " + str(end - beg) + str("sec."))
        #print("Added Linear Layers")
        # print()
        #print("Entering col2im")
        beg = time.time()
        col2im(self, out, H_prime, W_prime, 1, self.outputs)
        end = time.time()
        #print("Exiting im2col after " + str(end - beg) + str("sec."))
        #print("Exiting col2im")


if __name__ == '__main__':
    weight = np.asarray([0, 1, 1, 1])
    inputs = enc_vec(name='nn_input', nb=4)
    outputs = enc_vec(name='nn_outputs', nb=4)
    layer = linear_layer_1d(name='linear', weight=weight,
                            inputs=inputs, outputs=outputs)
    print(layer)



'''
All the code in this file has been copied from
the original SHEEP package. This is necessary to make the
matSHEEP package easy to use.
'''

import uuid
import os
import re


if "SHEEP_HOME" in os.environ.keys():
    BASE_DIR = os.environ["SHEEP_HOME"]
else:
    BASE_DIR = os.environ["HOME"] + "/SHEEP"

CIRCUIT_DIR_MID = BASE_DIR + "/benchmark_inputs/mid_level/circuits/"
INPUTS_DIR_MID = BASE_DIR + "/benchmark_inputs/mid_level/inputs/TMP/"
EXECUTABLE_DIR = BASE_DIR + "/build/bin"

TMP_INPUTS_DIR = BASE_DIR + "/benchmark_inputs/mid_level/inputs/TMP"
if not os.path.exists(TMP_INPUTS_DIR):
    os.system("mkdir " + TMP_INPUTS_DIR)


TMP_PARAMS_DIR = BASE_DIR + "/benchmark_inputs/params/TMP"
if not os.path.exists(TMP_PARAMS_DIR):
    os.system("mkdir " + TMP_PARAMS_DIR)


def write_inputs_file(value_dict):
    """
    write k,v pairs into a file.  Randomly generate the filename
    and return to the user.
    """
    filename = TMP_INPUTS_DIR + "/inputs-" + str(uuid.uuid4()) + ".inputs"
    inputs_file = open(filename, "w")
    for k, v in value_dict.items():
        inputs_file.write(k + " " + str(v) + "\n")
    inputs_file.close()
    return filename


def write_params_file(param_dict):
    """
    write k,v pairs into a file.  Randomly generate the filename
    and return to the user.
    """
    filename = TMP_PARAMS_DIR + "/params-" + str(uuid.uuid4()) + ".params"
    params_file = open(filename, "w")
    for k, v in param_dict.items():
        params_file.write(k + " " + str(v) + "\n")
    params_file.close()
    return filename


def cleanup_time_string(t):
    """
    convert from microseconds to seconds, and only output 3 s.f.
    """
    time_in_seconds = float(t) / 1e6

    if time_in_seconds < 1:
        time_in_seconds = round(time_in_seconds, 5)
    else:
        time_in_seconds = int(time_in_seconds)
    timestring = str(time_in_seconds)
    return timestring


def parse_test_output(outputstring, debug_filename=None):
    """
    Extract values from the stdout output of the "benchmark" executable.
    return a dict in the format { "processing times (seconds)" : {},
    "outputs" : {}, "sizes" : {}, "params":{}, "sizes"{} }
    """
    results = {}
    processing_times = {}
    test_outputs = {}
    params = {}
    sizes = {}
    clear_check = {}
    in_results_section = False
    in_processing_times = False
    in_outputs = False
    if debug_filename:
        debugfile = open(debug_filename, "w")
# parse the file, assuming we have processing times then outputs.
    for line in outputstring.decode("utf-8").splitlines():
        if debug_filename:
            debugfile.write(line + "\n")
# read lines where parameters are printed out
        param_search = re.search("Parameter ([\S]+) = ([\d]+)", line)
        if param_search:
            params[param_search.groups()[0]] = param_search.groups()[1]
# read lines where sizes of keys or ciphertexts are printed out
        size_search = re.search("size of ([\S]+):[\s]+([\d]+)", line)
        if size_search:
            sizes[size_search.groups()[0]] = size_search.groups()[1]
        if in_results_section:
            # parse the check against clear context:
            if line.startswith("Cleartext check"):
                clear_check["is_correct"] = "passed OK" in line

            if in_processing_times:
                num_search = re.search("([\w]+)\:[\s]+([\d][\d\.e\+]+)", line)
                if num_search:
                    label = num_search.groups()[0]
                    processing_time = num_search.groups()[1]
                    processing_time = cleanup_time_string(
                        processing_time)  # and convert to seconds
                    processing_times[label] = processing_time
                if "Output values" in line:
                    in_processing_times = False
                    in_outputs = True
            elif in_outputs:
                output_search = re.search("([\w]+)\:[\s]+([-\d]+)", line)
                if output_search:
                    label = output_search.groups()[0]
                    val = output_search.groups()[1]
                    test_outputs[label] = val
                if "END RESULTS" in line:
                    in_results_section = False
            elif "Processing times" in line:
                in_outputs = False
                in_processing_times = True
        elif "=== RESULTS" in line:
            in_results_section = True
            pass
    results["Processing times (s)"] = processing_times
    results["Outputs"] = test_outputs
    results["Object sizes (bytes)"] = sizes
    results["Parameter values"] = params
    results["Cleartext check"] = clear_check
    if debug_filename:
        debugfile.close()
    return results



from distutils.core import setup

setup(
    name='matSHEEP',
    version='1.4.4',
    author='Amartya Sanyal',
    author_email='amartya18x@gmail.com',
    packages=['matSHEEP'],
    #scripts=['bin/stowe-towels.py', 'bin/wash-towels.py'],
    # url='http://pypi.python.org/pypi/TowelStuff/',
    license='LICENSE.txt',
    description='Programmatic interface to SHEEP',
    long_description=open('README.txt').read(),
    install_requires=[
        "numpy",
        "multiprocessing",
        "networkx",
        "tqdm"
    ],
)



from particles import variables, enc_vec
from interactions import mono_assign, bi_assign, mini_mod, tri_assign
import copy


class oneb_adder(mini_mod):
    def __init__(self, name, inputs, outputs, nb=None,
                 randomize_temps=1, carry=True):
        self.randomize_temps = randomize_temps
        mini_mod.__init__(self, name, inputs, outputs)
        if carry:
            assert(len(inputs) == 3), "You need to have 3 inputs"
        else:
            assert(len(inputs) == 2), "You need to have 2 inputs"
        self.nb = nb
        if self.nb is not None:
            assert(self.nb == 1)
        self.create(carry)

    def create(self, carry):
        if carry:
            cin = self.inputs[0]

        x = self.inputs[carry]
        y = self.inputs[carry + 1]
        temp1 = variables(self.name + '_tmp1',
                          randomize_name=self.randomize_temps)
        temp2 = variables(self.name + '_tmp2',
                          randomize_name=self.randomize_temps)
        temp3 = variables(self.name + '_tmp3',
                          randomize_name=self.randomize_temps)
        self.add(bi_assign(ass_type='xor', lhs=[x, y], rhs=temp1))

        if carry:
            self.add(bi_assign(ass_type='xor', lhs=[temp1, cin],
                               rhs=self.outputs[0]))
        else:
            self.add(mono_assign(ass_type='alias', lhs=[temp1],
                                 rhs=self.outputs[0]))

        self.add(bi_assign(ass_type='and', lhs=[x, y],
                           rhs=temp2))

        if carry:
            self.add(bi_assign(ass_type='and', lhs=[cin, temp1],
                               rhs=temp3))
            self.add(bi_assign(ass_type='or', lhs=[temp3, temp2],
                               rhs=self.outputs[1]))
        else:
            self.add(mono_assign(ass_type='alias', lhs=[temp2],
                                 rhs=self.outputs[1]))


class nb_adder(mini_mod):
    def __init__(self, name, inputs, outputs, nb, randomize_temps=2,
                 carry_out=True, carry_in=False):
        '''
        Adds x1 and x2 where the bit length of both is nb
        Inputs will be structured as [cin x1 x2] - size(2*nb)+1
        Outputs - size (b+1)
        '''
        mini_mod.__init__(self, name, inputs, outputs)
        self.randomize_temps = randomize_temps
        self.nb = nb
        if not carry_in:
            assert(self.nb * 2 == len(inputs)), "You need to have " + \
                str(2 * self.nb) + " inputs not " + str(len(self.inputs))
        else:
            assert(self.nb * 2 + 1 == len(inputs)), "You need to have " + \
                str(2 * self.nb + 1) + " inputs not " + str(len(self.inputs))

        self.carry_list = [variables(self.name + '_carry_' + str(i),
                                     randomize_name=self.randomize_temps)
                           for i in range(self.nb + 1)]

        # TODO - Fill with zeros
        assert(len(self.outputs) ==
               self.nb + 1), "Output bitwidth should be length of adder: " + \
            str(self.nb + 1) + " not " + str(len(self.outputs))
        self.create(carry_out=carry_out, carry_in=carry_in)

    def create(self, carry_out, carry_in):
        inp1 = self.inputs[0 + int(carry_in):self.nb + int(carry_in)]
        inp2 = self.inputs[self.nb + int(carry_in):]

        if carry_in:
            self.add(mono_assign(ass_type='alias',
                                 lhs=[self.inputs[0]],
                                 rhs=self.carry_list[0]))
            self.carry_list[0] = self.inputs[0]
        for bit_id in range(self.nb):
            if bit_id > 0 or carry_in:
                self.add(oneb_adder(name=self.name + '2b_adder_' + str(bit_id),
                                    inputs=[self.carry_list[bit_id],
                                            inp1[bit_id],
                                            inp2[bit_id]],
                                    outputs=[self.outputs[bit_id],
                                             self.carry_list[bit_id + 1]],
                                    carry=True))
            else:
                self.add(oneb_adder(name=self.name + '2b_adder_' + str(bit_id),
                                    inputs=[inp1[bit_id],
                                            inp2[bit_id]],
                                    outputs=[self.outputs[bit_id],
                                             self.carry_list[bit_id + 1]],
                                    carry=bit_id > 0))
        if carry_out:
            self.add(mono_assign(ass_type='alias',
                                 lhs=[self.carry_list[-1]],
                                 rhs=self.outputs[-1]))


class nb_adder_xy(mini_mod):
    '''
    Adds two number of different bit lengths
    '''

    def __init__(self, name, inputs, outputs, nb, randomize_temps=2,
                 carry_out=True):
        '''
        Adds x1 and x2 where the bit length of both is nb
        Inputs will be structured as [cin x1 x2] - size(2*nb)+1
        Outputs - size (b+1)
        '''
        self.x = inputs[0]
        self.y = inputs[1]
        mini_mod.__init__(self, name, self.x + self.y, outputs)

        self.len1 = len(self.x)
        self.len2 = len(self.y)

        if self.len1 < self.len2:
            tmp = copy.deepcopy(self.x)
            self.x = copy.deepcopy(self.y)
            self.y = tmp

        self.randomize_temps = randomize_temps
        self.nb = nb
        assert(self.nb == self.len1), "You need to have " + \
            str(self.nb) + " inputs not " + str(len(self.x))

        self.carry_list = [variables(self.name + '_carry_' + str(i),
                                     randomize_name=self.randomize_temps)
                           for i in range(self.nb + 1)]

        assert(len(self.outputs) ==
               self.nb + 1), "Output bitwidth should be length of adder: " + \
            str(self.nb + 1) + " not " + str(len(self.outputs))
        self.create(carry_out=carry_out)

    def create(self, carry_out):
        inp1 = self.x
        inp2 = self.y

        for bit_id in range(self.nb):
            if bit_id > 0 and bit_id < self.len2:
                self.add(oneb_adder(name=self.name + '2b_adder_' + str(bit_id),
                                    inputs=[self.carry_list[bit_id],
                                            inp1[bit_id],
                                            inp2[bit_id]],
                                    outputs=[self.outputs[bit_id],
                                             self.carry_list[bit_id + 1]],
                                    carry=True))
            elif bit_id == 0:
                self.add(oneb_adder(name=self.name + '2b_adder_' + str(bit_id),
                                    inputs=[inp1[bit_id],
                                            inp2[bit_id]],
                                    outputs=[self.outputs[bit_id],
                                             self.carry_list[bit_id + 1]],
                                    carry=bit_id > 0))
            elif bit_id >= self.len2:
                self.add(oneb_adder(name=self.name + '2b_adder_' + str(bit_id),
                                    inputs=[self.carry_list[bit_id],
                                            inp1[bit_id]],
                                    outputs=[self.outputs[bit_id],
                                             self.carry_list[bit_id + 1]],
                                    carry=False))
        if carry_out:
            self.add(mono_assign(ass_type='alias',
                                 lhs=[self.carry_list[-1]],
                                 rhs=self.outputs[-1]))


class compare_cp(mini_mod):
    def __init__(self, name, inputs, outputs):
        mini_mod.__init__(self, inputs=inputs[0], outputs=outputs, name=name)
        self.plaintext = [x for x in bin(inputs[1])[::-1][:-2]]
        self.create()

    def create(self):
        if isinstance(self.outputs, enc_vec):
            rel_out = self.outputs[0]
        else:
            rel_out = self.outputs
        temp1 = enc_vec(name=self.name + 'out_bit', nb=len(self.inputs))
        self.add(bi_assign(ass_type='and', lhs=[self.inputs[0],
                                                self.zero],
                           rhs=temp1[0]))
        temp2 = enc_vec(name=self.name + 'flag_bit', nb=len(self.inputs))
        for idx in range(len(self.inputs)):
            if idx < len(self.plaintext) and int(self.plaintext[idx]) == 1:
                self.add(tri_assign(ass_type='mux',
                                    lhs=[self.inputs[idx],
                                         temp1[idx],
                                         self.zero],
                                    rhs=temp2[idx]))
            else:
                self.add(tri_assign(ass_type='mux',
                                    lhs=[self.inputs[idx],
                                         self.one,
                                         temp1[idx]],
                                    rhs=temp2[idx]))
            if idx < len(self.inputs) - 1:
                self.add(mono_assign(ass_type='alias',
                                     lhs=[temp2[idx]], rhs=temp1[idx + 1]))
            else:
                self.add(mono_assign(ass_type='alias',
                                     lhs=[temp2[idx]], rhs=rel_out))


if __name__ == '__main__':
    x = variables('x')
    y = variables('y')
    cin = variables('cin')
    s = variables('sum')
    c = variables('carry')
    b2_adder = oneb_adder('2b_adder', [cin, x, y], [s, c], 1)
    print(b2_adder)
